/* App Grid */

$app_icon_size: 96px;

// app icons
.icon-grid {
  row-spacing: $base_padding * 2;
  column-spacing: $base_padding * 2;
  max-row-spacing: $base_padding * 7;
  max-column-spacing: $base_padding * 7;
  page-padding-top: $base_padding * 4;
  page-padding-bottom: $base_padding * 4;
  page-padding-left: $base_padding * 3;
  page-padding-right: $base_padding * 3;
}

/* App Icons */

// Icon tiles in the app grid
.app-well-app {
  @include overview_icon($system_fg_color);

  .overview-icon {
    padding: $base_padding*2;
    border-radius: $base_border_radius*3;
    color: $system_fg_color;
  }
  .overview-icon.overview-icon-with-label {
    > StBoxLayout {
      spacing: $base_padding;
    }
  }
}

// app folders
.app-well-app.app-folder {
  @include overview_icon($system_fg_color, $flat: false);
}

// expanded folder
.app-folder-dialog {
  border-radius: $modal_radius*4;
  background-color: $system_overlay_bg_color;

  @if $is_highcontrast {
    padding: $base_padding*2 2px $base_padding*2 2px;
    box-shadow: inset 0 0 0 2px $hc_inset_color;
  } @else {
    padding: $base_padding*2 1px $base_padding*2 1px;
    box-shadow: inset 0 0 0 1px $system_borders_color;
  }

  & .folder-name-container {
    padding: 24px 36px 0;
    spacing: 12px;

    & .folder-name-label,
    & .folder-name-entry {
      @extend %title_1;
    }

    & .folder-name-entry {
      width: 12em;
      color: $system_fg_color !important;
    }

    /* FIXME: this is to keep the label in sync with the entry */
    & .folder-name-label { padding: 5px 7px; color: $system_fg_color; }

    & .edit-folder-button {
      @extend %button;
      padding: 0;
      width: 36px;
      height: 36px;
      border-radius: 99px;
      & > StIcon { icon-size: $base_icon_size }
    }
  }

  & .icon-grid {
    row-spacing: $base_padding * 2;
    column-spacing: $base_padding * 5;
    page-padding-top: 0;
    page-padding-bottom: 0;
    page-padding-left: 0;
    page-padding-right: 0;
  }

  & .page-indicators {
    margin-bottom: 18px;
  }
}

.app-folder-dialog-container {
  padding: $base_padding*2;
  width: 720px;
  height: 720px;
}

// Running app indicator (also shown in dash)
.app-well-app-running-dot {
  height: 5px;
  width: 5px;
  border-radius:5px;
  margin-bottom: 8px;
  background-color: $system_fg_color;
}

// Rename popup for app folders
.rename-folder-popup {
  .rename-folder-popup-item {
    spacing: $base_padding;
    &:ltr, &:rtl { padding: 0 $base_padding * 2; }
  }
}

// App Grid pagination indicators
.page-indicator {
  padding: $base_padding $base_padding * 2 0;
  transition-duration:400ms;

  .page-indicator-icon {
    width: 10px;
    height: 10px;
    border-radius: 10px; // the same as height&width
    background-color: white;
    transition-duration: 400ms;
  }
}

.apps-scroll-view {
  padding: 0;
}

// shutdown and other actions in the grid
.system-action-icon {
  background-color: transparentize($system_fg_color,.9);
  color: $system_fg_color;
  border-radius: 99px;
  icon-size: $app_icon_size * 0.5;
  @if $is_highcontrast {
    box-shadow: inset 0 0 0 2px $hc_inset_color;
  }
}

.page-navigation-hint {
  &.dnd {
    background: rgba(255, 255, 255, 0.1);
  }

  &.next:ltr,
  &.previous:rtl {
    background-gradient-start: rgba(255, 255, 255, 0.05);
    background-gradient-end: transparent;
    background-gradient-direction: horizontal;
    border-radius: $modal_radius*1.5 0px 0px $modal_radius*1.5;
  }

  &.previous:ltr,
  &.next:rtl {
    background-gradient-start: transparent;
    background-gradient-end: rgba(255, 255, 255, 0.05);
    background-gradient-direction: horizontal;
    border-radius: 0px $modal_radius*1.5 $modal_radius*1.5 0px;
  }
}

.page-navigation-arrow {
  margin: 6px;
  padding: 18px;
  width: 24px;
  height: 24px;
  border-radius: 99px;
  transition-duration: 100ms;

  > StIcon { color: $system_fg_color;}

  @if $is_highcontrast {
    @include button(normal, $system_fg_color, transparentize($system_bg_color, 0.5));
  }

  &:insensitive { @include button(undecorated, $system_fg_color, transparentize($system_bg_color, 0.5));}
  &:hover { @include button(hover, $system_fg_color, transparentize($system_bg_color, 0.5));}
  &:active { @include button(active, $system_fg_color, transparentize($system_bg_color, 0.5));}
}
